
use master
go
create database simple
go
alter database simple add filegroup simple_mod contains memory_optimized_data
go
-- adapt filename as needed
alter database simple add file (name='simple_mod', filename='c:\data\simple_mod') to filegroup simple_mod
go

use simple
go
create table dbo.t1
(c1 int not null primary key nonclustered hash with (bucket_count=10000),
 c2 varchar(10))
with (memory_optimized=on)
go

insert dbo.t1 values(1, 'Ann'),(2, 'Bob'),(3, 'Cathy'),(4, 'Dot')
go
delete dbo.t1 where c1 < 3
select * from dbo.t1

create table dbo.t2
(c1 int not null primary key nonclustered hash with (bucket_count=10000),
 c2 varchar(10))
with (memory_optimized=on)
go

insert dbo.t2 values(1, 'Ann'),(2, 'Bob'),(3, 'Cathy'),(4, 'Dot')
go

-- transaction isolation level

-- The database simple SNAPSHOT and RCSI turned off at the database level
-- I'm running at default isolation level of read committed
-- XACT_ABORT session setting is turned off
select * from sys.databases
dbcc useroptions
select @@trancount -- (for your convenience while running the demos)

--1. SNAPSHOT
begin tran
select * from dbo.t1 with (snapshot)
-- update a row in another session
commit -- OK
go

begin tran
select * from dbo.t1 with (snapshot)
-- update a row in another session
select * from dbo.t1 with (snapshot) -- sees old value
commit -- OK

--2. REPEATABLE READ
begin tran
select * from dbo.t1 with (repeatableread)
-- update a row in another session
commit -- Msg 41305, Level 16, State 0, Line 39 The current transaction failed to commit due to a repeatable read validation failure.

begin tran
select * from dbo.t1 with (repeatableread)
-- insert a row in another session
commit -- OK

begin tran
select * from dbo.t1 with (repeatableread)
-- insert a row in another session
select * from dbo.t1 with (repeatableread) -- does not see new row
commit -- OK

begin tran
-- update row in another session before reading
select * from dbo.t1 with (repeatableread) -- sees new value
commit -- OK

begin tran
select * from dbo.t2 with (repeatableread) -- transaction starts with first SQL statement
-- update row in t1 in another session before reading
select * from dbo.t1 with (repeatableread)
commit -- Msg 41305, Level 16, State 0, Line 72 The current transaction failed to commit due to a repeatable read validation failure.

-- 3. SERIALIZABLE
begin tran
select * from dbo.t1 with (serializable)
-- insert a row in another session
commit -- OK Msg 41325, Level 16, State 0, Line 55 The current transaction failed to commit due to a serializable validation failure.

-- More stuff...

begin tran
select * from dbo.t1 with (repeatableread)
-- insert a row in another session
select * from dbo.t1 with (serializable) -- does not see new row
commit -- OK  Msg 41325, Level 16, State 0, Line 61 The current transaction failed to commit due to a serializable validation failure.

begin tran
insert dbo.t1 with (snapshot) values(5, 'Fred')
-- insert row with same primary key in another session and commit other session (this works)
commit -- Msg 41325, Level 16, State 0, Line 66 The current transaction failed to commit due to a serializable validation failure.

begin tran
select * from dbo.t1 with (readcommitted) -- the table option 'readcommitted' is not supported with memory optimized tables.
commit -- OK
go

begin tran
select * from dbo.t1 -- Msg 41368, Level 16, State 0, Line 37 Accessing memory optimized tables using the READ COMMITTED isolation level is supported only for autocommit transactions. It is not supported for explicit or implicit transactions. Provide a supported isolation level for the memory optimized table using a table hint, such as WITH (SNAPSHOT).
commit -- no corresponding transaction

select @@trancount

begin tran
-- update row 2 in another transaction and commit
update dbo.t1 with (snapshot) set c2 = 'Sam' where c1 = 2 -- tx starts here, 2 is already updated/commited to new value
commit -- OK

begin tran
select * from dbo.t2 with (snapshot) -- starts the transaction
-- update row 2 in another transaction and commit
update dbo.t1 with (snapshot) set c2 = 'Sam' where c1 = 2
-- Msg 41302, Level 16, State 110, Line 120
-- The current transaction attempted to update a record that has been updated since this transaction started. The transaction was aborted.
-- Msg 3998, Level 16, State 1, Line 120
-- Uncommittable transaction is detected at the end of the batch. The transaction is rolled back.
commit -- transaction already rolled back (above)

begin tran
select * from dbo.t2 with (snapshot) -- starts the transaction
-- update row 2 in another transaction and commit
delete dbo.t1 with (snapshot) where c1 = 2
-- Msg 41302, Level 16, State 111, Line 130
-- The current transaction attempted to update a record that has been updated since this transaction started. The transaction was aborted.
-- Msg 3998, Level 16, State 1, Line 130
-- Uncommittable transaction is detected at the end of the batch. The transaction is rolled back.
-- The statement has been terminated.
commit -- transaction already rolled back (above)

begin tran
select * from dbo.t2 with (snapshot) -- starts the transaction
select * from dbo.t1 with (snapshot)
-- drop table t1 in the other session
commit -- OK

begin tran
select * from dbo.t1 with (snapshot)
-- drop table t1 in the other session
select * from dbo.t1 with (snapshot) -- invalid object name dbo.t1
commit -- OK

begin tran
select * from dbo.t1 with (repeatableread)
-- drop table t1 in the other session
select * from dbo.t1 with (repeatableread) -- OK
commit -- OK, (and drop works after this tx committed)

begin tran
insert into dbo.t1 with (snapshot) values(5, 'Fred')
-- drop table t1 in the other session - the drop statement hangs
commit -- OK, (and drop works after this tx committed)
go

create procedure dbo.select_t1
with native_compilation, schemabinding, execute as owner
as
begin atomic
with (transaction isolation level=snapshot, language=N'us_english')
select c1, c2 from dbo.t1
end
go

create procedure dbo.insert_t1(@c1 int, @c2 varchar(10))
with native_compilation, schemabinding, execute as owner
as
begin atomic
with (transaction isolation level=snapshot, language=N'us_english')
insert into dbo.t1 values(@c1, @c2)
end
go

create procedure dbo.select_t1_rr
with native_compilation, schemabinding, execute as owner
as
begin atomic
with (transaction isolation level=repeatable read, language=N'us_english')
select c1, c2 from dbo.t1
end
go

create procedure dbo.insert_t1_rr(@c1 int, @c2 varchar(10))
with native_compilation, schemabinding, execute as owner
as
begin atomic
with (transaction isolation level=repeatable read, language=N'us_english')
insert into dbo.t1 values(@c1, @c2)
end
go

begin tran
execute dbo.select_t1
-- drop table t1 in the other session -- Cannot DROP TABLE 't1' because it is being referenced by object 'select_t1'.
commit -- OK, (and drop works after this tx committed)
go

begin tran
execute dbo.select_t1
-- drop procedure dbo.select_t1 table t1 in the other session (works 
commit -- OK
go

drop procedure dbo.select_t1
drop procedure dbo.select_t1_rr
drop procedure dbo.insert_t1_rr

-- mistake, but Msg 10773, Level 16, State 114, Procedure select_t1_rr, Line 214
-- The feature 'transactional DDL' is not yet implemented with natively compiled stored procedures.
-- can't create proc inside a transaction

create procedure dbo.select_t1_rr
with native_compilation, schemabinding, execute as owner
as
begin atomic
with (transaction isolation level=repeatable read, language=N'us_english')
select c1, c2 from dbo.t1
end
go

begin tran
execute dbo.select_t1_rr
-- drop procedure dbo.select_t1_rr and table t1 in the other session (works)
commit -- Msg 41305, Level 16, State 0, Line 229 The current transaction failed to commit due to a repeatable read validation failure.
go

begin tran
execute dbo.insert_t1 5, 'Fred'
-- drop procedure and table t1 in the other session
commit -- Msg 41305, Level 16, State 0, Line 211 The current transaction failed to commit due to a repeatable read validation failure.

select @@trancount
rollback

begin tran
insert dbo.t1 with (snapshot) values(5, 'Fred')
-- drop table t1 in the other session, waits
commit

begin tran
-- select * from dbo.t2 with (snapshot)
-- update same row in other session, but dont commit
update dbo.t1 with (snapshot) set c2 = 'Mary' where c1 = 2  -- this fails